from __future__ import division
from collections import namedtuple
from functools import partial
from math import log, exp

# A collection of common bijections. Bijections can be composed by addition: (f
# + g)(x) = (g.f)(x) or something like that. I can't remember. It seemed good
# way at the time!

_Bijection = namedtuple('Bijection', ('a', 'b', 'ab', 'ba'))


class Bijection(_Bijection):
    @property
    def inverse(self):
        return Bijection(a=self.b, b=self.a,
                         ab=self.ba, ba=self.ab)

    def __add__(a, b):
        return compose(a, b)

    def __call__(self, x):
        return self.ab(x)


def _composed_f(f, g, x):
    return g(f(x))


def f_compose(f, g):
    return partial(_composed_f, f, g)


def compose(f, g):
    if f.b != g.a:
        raise RuntimeError('Codomain of f must be the domain of g')

    return Bijection(a=f.a,
                     b=g.b,
                     ab=f_compose(f.ab, g.ab),
                     ba=f_compose(g.ba, f.ba))


# Linear spaces
def linspace_ab(a, m, x):
    return a + m * x


def linspace_ba(a, r, x):
    return r * (x - a)


def linspace(u, v, a='', b=''):
    m = v - u

    ab = partial(linspace_ab, u, m)
    ba = partial(linspace_ba, u, 1.0 / m)

    return Bijection(a=a, b=b, ab=ab, ba=ba)


# Logarithmic spaces
def logspace_ab(y0, b, x):
    return y0 * exp(b * x)


def logspace_ba(iy0, ib, x):
    return ib * log(iy0 * x)


def logspace(u, v, a='', b=''):
    nb = log(v / u)
    ib = 1.0 / nb
    y0 = u
    iy0 = 1.0 / y0

    ab = partial(logspace_ab, y0, nb)
    ba = partial(logspace_ba, iy0, ib)

    return Bijection(a=a, b=b, ab=ab, ba=ba)


# Combinations
def separable_ab(fx, fy, (x, y)):
    return fx(x), fy(y)


def separable(x, y, a=None, b=None):
    ab = partial(separable_ab, x.ab, y.ab)
    ba = partial(separable_ab, x.ba, y.ba)
    a = a or (x.a, y.a)
    b = b or (x.b, y.b)
    return Bijection(ab=ab, ba=ba, a=a, b=b)


# Convenience constructors
def bilin((ux, uy), (vx, vy), a=None, b=None):
    return separable(linspace(ux, vx),
                     linspace(uy, vy),
                     a=a, b=b)


def bilog((ux, uy), (vx, vy), a=None, b=None):
    return separable(logspace(ux, vx),
                     logspace(uy, vy),
                     a=a, b=b)

# TODO: optimisations, if they're necessary. This is one of the possible
# bottlenecks, but it may turn out not to be; especially if full-stack
# transformations are relatively rare in plotting. There are a few nice
# things we can do, like combine runs of log and linear transformations
# (bunching), but whether they'll actually beat just writing an optional
# swapout module in Cython is dubious at best. I'll be placing bets on
# Cython. :)

#a = bilin((0, 0), (100, 100))
#b = bilog((10, 10), (1000, 1000))
#c = bilin((0, 1), (50, 51))
#
#d = a.inverse + b.inverse + c
#
#from random import random
#
#for i in xrange(10000):
#    x = random()
#    y = random()
#    d((x, y))



from collections import namedtuple
from functools import partial
from operator import mul

# Units are distinct from spaces, but nonetheless important.

# Define our namedtuples, many of which will be bases
Unit = namedtuple('Unit', ('name', 'to_pt', 'from_pt'))
_Coord = namedtuple('Coord', ('point', 'units', 'unit'))
_Point = namedtuple('Point', ('x', 'y'))


def unit(pts, name=''):
    return Unit(to_pt=partial(mul, pts),
                from_pt=partial(mul, 1.0 / pts),
                name=name)


DEFAULT_UNITS = dict(
    mm=unit(72.0 / 25.4, 'mm'),
    cm=unit(72.0 / 2.54, 'cm'),
    inch=unit(72.0, 'inch'),
    pt=unit(1.0, 'pt')
)
DEFAULT_UNIT = DEFAULT_UNITS['pt']


class Point(_Point):
    def __call__(self, x, y):
        return self._replace(x=x, y=y)

    @property
    def point(self):
        return self

    def __sub__(a, b):
        return Point(a.x - b.x, a.y - b.y)


class UnitChanger(object):
    """Change space"""
    __slots__ = 'unit', '__doc__'
    #func_name = __name__ = 'UnitChanger'

    @property
    def __doc__(self):
        return "Change unit to %s" % self.unit.name

    def __init__(self, unit):
        self.unit = unit

    def __get__(self, ins, cls):
        if ins:
            return ins.switch_unit(self.unit)
        else:
            return self

    def __call__(self, x, y):
        pass

    def __set__(self, *args):
        raise AttributeError('Unable to set UnitChanger')

    def __repr__(self):
        return "Change units to %s" % self.unit.name


class Coord(_Coord):
    __slots__ = ()

    @property
    def base(self):
        return self.pt.point

    def switch_unit(self, unit):
        g = self.unit.to_pt
        f = unit.from_pt
        x, y = self.point
        return self._replace(
            point=self.point._replace(x=f(g(x)), y=f(g(y))),
            unit=unit
        )

    def __new__(cls, x=0, y=0):
        return _Coord.__new__(cls, point=Point(x, y),
                              units=DEFAULT_UNITS,
                              unit=DEFAULT_UNIT)

    def __repr__(self):
        return ('Coord((%g, %g) %s)' %
                (self.point.x, self.point.y, self.unit.name))

    def __call__(self, x, y):
        if x is not None and y is not None:
            return self._replace(point=Point(x, y))
        elif x is not None:
            _, cy = self.point
            return self._replace(point=Point(x, cy))
        elif y is not None:
            cx, _ = self.point
            return self._replace(point=Point(cx, y))
        else:
            return self._replace(point=Point(0, 0))


# Set all the UnitChangers on the usual Coord
for i, j in DEFAULT_UNITS.iteritems():
    setattr(Coord, i, UnitChanger(j))


# A Raw coordinate class
class RawCoord(_Coord):
    __slots__ = ()
    
    @property
    def base(self):
        return self.point

    def switch_unit(self, unit):
        raise AttributeError('Cannot switch unit')

    def __new__(cls, x=0, y=0):
        return _Coord.__new__(cls, point=Point(x, y),
                              units=None,
                              unit=None)

    def __repr__(self):
        return 'Coord(%g, %g)' % self.point

    def __call__(self, x, y):
        if x is not None and y is not None:
            return self._replace(point=Point(x, y))
        elif x is not None:
            _, cy = self.point
            return self._replace(point=Point(x, cy))
        elif y is not None:
            cx, _ = self.point
            return self._replace(point=Point(cx, y))
        else:
            return self._replace(point=Point(0, 0))

# Thing is that this is now easily verifiable. We would like to expose
# some of this magic to the user of the cursors, and we can now do that
# by composition rather than it being something built (quite horribly)
# into the system But now, how do we deal with alternative spaces?  See
# how simple? :)



# External modules
import collections
import colorsys

# Immutable colour system based around namedtuples.

_RGB = collections.namedtuple("RGB", ('r', 'g', 'b'))
_HSL = collections.namedtuple("HSL", ('h', 's', 'l'))

class HSL(_HSL):
    @property
    def hsl(self):
        return self

    @property
    def rgb(self):
        h, s, l = self
        return RGB(*colorsys.hls_to_rgb(h, l, s))

class RGB(_RGB):
    @property
    def hsl(self):
        h, l, s = colorsys.rgb_to_hls(*self)
        return HSL(h, s, l)

    @property
    def rgb(self):
        return self

# A general interpolator. This means that we can avoid handling rgb, hsl,
# (cielab) interpolations as special cases and just get on with things nicely.

def interpolate(patch, val):
    pass

def make_interpolator(frm, to):
    partial(interpolate, patch)


